unit OOPHilit;

{  Electro-Concept Mauricie, 1999 }
{ Implements TOOPHighlighter object, used for Delphi like syntax highlighting
  in a TPlusMemo }

{ Note: to install this component on your palette, add file OOPReg in your package }

interface

uses Classes, Graphics, Controls, PlusMemo, ExtHilit, NbHilit;

const OOPDelimiters: TSetOfChar = ['"', ' ', '''', '(', ')', ',', '.', '/', ':', ';',
                                   '<', '>', '[', ']', '{', '}', #9, '=', '-'];
      OOPKeywordContext    = 1;
      PascalStringContext  = 2;
      OOPCommentContext    = 3;
      DirectiveContext     = 4;
      NumberContext        = 5;

      ASMContext           = 6;
      ASMKeywordContext    = 7;
      ASMStringContext     = 8;
      ASMCommentContext    = 3;

type
     TOOPKeywordList = class(TPersistent) { a class provided for the purpose of editing keyword
                                             formats and streaming from old (pre 4.1) .dfm files }
        protected
          ParentHighlighter: TExtHighlighter;
          procedure DefineProperties(Filer: TFiler); override;
          procedure WriteData(Writer: TWriter);
        end;

     TOOPHighlighter = class(TCustomExtHighlighter)
       private
         fKeywordsOOP,       fPreviousKeywordsOOP,
         fKeywordsASM,       fPreviousKeywordsASM,
         fAsmHighlight,      fStringsHilit,
         fCommentSingleLine, fCommentBrace,
         fCommentParAst,     fDirectives,
         fNumbers           : THighlightInfo;

         fKeywords          : TOOPKeywordList;
         fApplyHChange      : Boolean;

         fNumHilit          : TNumberHighlighter;

         procedure ReadOldFormats(Reader: TReader);
         procedure ReadNewFormats(Reader: TReader);
         procedure WriteFormats(Writer: TWriter);

         procedure HighlightChange(Sender: TObject);

         procedure setKeywordsOOP(h: THighlightInfo);
         procedure setKeywordsASM(h: THighlightInfo);
         procedure setASMHighlight(h: THighlightInfo);
         procedure setStringsHilit(h: THighlightInfo);
         procedure setCommentSingleLine(h: THighlightInfo);
         procedure setCommentBrace(h: THighlightInfo);
         procedure setCommentParAst(h: THighlightInfo);
         procedure setDirectives(h: THighlightInfo);
         procedure setNumbers(h: THighlightInfo);

         procedure GetKeywordConfirmation(Sender: TObject; StartKey, StopKey: TPlusNavigator; KIndex: Integer;
                                          var Accept: Boolean);
         procedure GetNumberConfirmation(Sender: TObject; Scope: SmallInt; var DoHilit: Boolean);

       protected
         procedure Loaded; override;
         procedure DefineProperties(Filer: TFiler); override;
         procedure ApplyKeywordsList(Start, Stop: TPlusNavigator; BaseIndex: Integer);  override;

       public
         constructor Create(AOwner: TComponent); override;
         destructor Destroy; override;

       published
         property KeywordsOOP   : THighlightInfo  read fKeywordsOOP   write setKeywordsOOP;
         property KeywordsASM   : THighlightInfo  read fKeywordsASM   write setKeywordsASM;
         property ASMHighlight  : THighlightInfo  read fASMHighlight  write setASMHighlight;
         property Strings       : THighlightInfo  read fStringsHilit  write setStringsHilit;
         property CommentSingleLine: THighlightInfo read fCommentSingleLine write setCommentSingleLine;
         property CommentBrace  : THighlightInfo  read fCommentBrace  write setCommentBrace;
         property CommentParAst : THighlightInfo  read fCommentParAst write setCommentParAst;
         property Directives    : THighlightInfo  read fDirectives    write setDirectives;
         property Numbers       : THighlightInfo  read fNumbers       write setNumbers;
         property Keywords      : TOOPKeywordList read fKeywords      write fKeywords;
         property SubHighlighter;  { promoted from TCustomExtHighlighter to allow nested highlighting, ex:
                                     url highlighter within comment sections }
       end;

implementation

uses SysUtils, Dialogs;

type sstring = {$IFDEF WIN32} string {$ELSE} string[20] {$ENDIF};

const
      OOPKeywordCount = 80;
      OOPKeywords: array[0..OOPKeywordCount-1] of sstring =
           ('and', 'array', 'as', 'begin',      { note: asm keyword must not be there... }
            'case', 'class', 'const', 'constructor',
            'default', 'destructor', 'dispinterface', 'div',
            'do', 'downto', 'dynamic', 'else',
            'end', 'except', 'exports', 'file',
            'finalization', 'finally', 'for', 'function',
            'goto', 'if', 'implementation', 'in',
            'inherited', 'initialization', 'inline', 'interface',
            'is', 'label', 'library', 'mod',
            'message', 'nil', 'not', 'object',
             'of', 'or', 'out', 'override',
            'packed', 'private', 'procedure', 'program',
            'property', 'protected', 'public', 'published',
            'raise', 'record', 'register', 'repeat',
            'resourcestring', 'set', 'shl', 'shr',
            'string', 'then', 'threadvar', 'to',
            'try', 'type', 'unit', 'until',
            'uses', 'var', 'virtual', 'while',
            'with', 'xor',
            'implements', 'index', 'nodefault', 'read',
            'stored', 'write');   { these last six highlighted only inside of property section }

      PropertyKeywordsIndex = 74;    { index of keywords highlighted only in property sections }

      ASMKeywordCount = 166;
      ASMKeywords: array[0..ASMKeywordCount-1] of sstring =
           (  'ASM', 'END',
              'AAA', 'AAS', 'AAD', 'AAM',         'ADC', 'ADD', 'AND', 'BOUND',
              'BSF', 'BSR', 'BT', 'BTC',         'BTR', 'BTS', 'CALL', 'CMP',
              'CBW', 'CWD', 'CDQ',  'CWDE',      'CLC', 'CLD', 'CLI',  'CMC',
              'CMPSB', 'CMPSD', 'CMPSW', 'DAA',  'DAS', 'DEC', 'DIV', 'ENTER',
              'HLT', 'IDIV', 'IMUL', 'IN',       'INC', 'INSB', 'INSW', 'INSD',
              'INT', 'INTO', 'IRET', 'IRETD',    'JA',  'JAE', 'JBE',  'JC',
              'JE',  'JCXZ','JG',  'JGE',        'JMP', 'JO',  'JNA', 'JNO',
              'JB',  'JNAE', 'JNB', 'JNC',       'JZ',  'JNE',  'JNZ', 'JNBE',
              'JS', 'JNS',   'JP',  'JPE',       'JNP', 'JPO',  'JL',  'JNGE',
              'JNL', 'JLE', 'JNG', 'JNLE',       'JECXZ', 'LAHF', 'LEA', 'LEAVE',
              'LDS', 'LES', 'LFS', 'LGS',        'LODSB', 'LODSW', 'LODSD', 'LOOP',
              'LOCK',   'REP',   'REPE', 'REPZ', 'REPNE',  'REPNZ', 'SEGES',  'SEGCS',
              'SEGSS',  'SEGDS', 'SEGFS','SEGGS', 'LOOPE', 'LOOPZ', 'LOOPNE', 'LOOPNZ',
              'LOOPD', 'LOOPDE', 'LOOPDZ', 'LOOPDNE','LOOPDNZ', 'MOV', 'MOVSX', 'MOVZX',
              'MOVSB', 'MOVSW', 'MOVSD', 'MUL',   'NEG', 'NOP', 'NOT', 'OR',
              'OUT', 'OUTSB', 'OUTSW', 'OUTSD',   'POP', 'POPF', 'POPA', 'POPAD',
              'POPFD', 'PUSH', 'PUSHF', 'PUSHA',  'PUSHAD', 'PUSHFD', 'RET', 'RETN',
              'RETF', 'SUB', 'SBB', 'RCL',        'RCR', 'ROL', 'ROR', 'SAL',
              'SHL', 'SAR', 'SHR', 'SHLD',        'SHRD', 'SAHF', 'SCASB', 'SCASW',
              'SCASD', 'STC', 'STD', 'STI',       'STOSB', 'STOSW', 'STOSD', 'TEST',
              'WAIT', 'XCHG', 'XLAT', 'XOR');

      AsmKIndex = OOPKeywordCount;
      AsmSSIndex = 11;    { index of asm ... end in StartStopKeys list }

function ConfirmKeywords(newHighlight, oldHighlight: THighlightInfo; kkind: string): Boolean;
var msg: string; savedOnChange: TNotifyEvent;
begin
  Result:= True;
  msg:= 'Are you sure you want to change ';
  if (newHighlight.AltFont<>oldHighlight.AltFont) or (newHighlight.Style<>oldHighlight.Style) then
    msg:= msg + 'the style '
  else
    if (newHighlight.Background<>oldHighlight.Background) or
       (newHighlight.Foreground<>oldHighlight.Foreground) then
         msg:= msg + 'the color '
    else Exit;

  Result:= MessageDlg(msg+ 'of all your '+kkind+' keywords?', mtConfirmation,  [mbOk, mbCancel], 0)=mrOK;
  if not Result then
    with newHighlight do
      begin
        savedOnChange:= OnChange;
        OnChange:= nil;
        Assign(oldHighlight);
        OnChange:= savedOnChange
      end
end;

constructor TOOPHighlighter.Create(AOwner: TComponent);
  procedure CreateHighlight(var h: THighlightInfo; style: TExtFontStyles; bk, fr: TColor);
    var newstyle: TPlusFontStyles;
    begin
      h:= THighlightInfo.Create;
      h.AltFont:=  TPlusFontStyle(fsAltFont) in TPlusFontStyles(style);
      newstyle:= TPlusFontStyles(style) - [TPlusFontStyle(fsAltFont)];
      h.Style:= TFontStyles(newstyle);
      h.Background:= bk;
      h.Foreground:= fr;
      h.OnChange:= HighlightChange
    end;

  var i: Integer;
  begin
    inherited Create(AOwner);

    fKeywords:= TOOPKeywordList.Create;
    fKeywords.ParentHighlighter:= TExtHighlighter(Self);

    OnKeyword:= GetKeywordConfirmation;  { special treatment for asm, index, read, write keywords }
    Delimiters:= OOPDelimiters;

    { Install keywords:
      Keywords contains first the Pascal keywords, followed by Assembler keywords }

    inherited Keywords.BeginUpdate;
    for i:= 0 to OOPKeyWordCount-1 do
      inherited Keywords.AddExtKeyword(OOPKeyWords[i], [woWholeWordsOnly], [fsBold],  OOPKeywordContext, 0, 0,
                                       crDefault, -1, -1);

    for i:= 0 to ASMKeywordCount-1 do
      inherited Keywords.AddExtKeyword(ASMKeywords[i], [woWholeWordsOnly], [fsBold], ASMKeywordContext, ASMContext, 0,
                                      crDefault, clYellow, -1);
    inherited Keywords.EndUpdate;

    { Install start-stop keys }
    StartStopKeys.AddExStartStopKey('property', ';', [woWholeWordsOnly], [], 0, 0, 0, crDefault,
                                          -1, -1, False);

    StartStopKeys.AddExStartStopKey('''', '''', [], [fsItalic], PascalStringContext, 0, 0, crDefault,
                                          clSilver, clBlack, True);

    StartStopKeys.AddExStartStopKey('//', '', [], [fsItalic], OOPCommentContext, 0, 0, crDefault,
                                          -1, clBlue, True);

    StartStopKeys.AddExStartStopKey('{$', '}', [], [fsItalic], DirectiveContext, 0, 0, crDefault,
                                          clTeal, clWhite, False);

    StartStopKeys.AddExStartStopKey('{', '}', [], [fsItalic], OOPCommentContext, 0,0, crDefault,
                                          -1, clNavy, False);

    StartStopKeys.AddExStartStopKey('(*', '*)', [], [fsItalic], OOPCommentContext, 0, 0, crDefault,
                                          -1, clGreen, False);

    StartStopKeys.AddExStartStopKey('''', '''', [], [fsItalic], ASMStringContext, ASMContext,0, crDefault,
                                          clSilver, clBlack, True);

    StartStopKeys.AddExStartStopKey('//', '', [], [fsItalic], ASMCommentContext, ASMContext, 0, crDefault,
                                          -1, clBlue, True);

    StartStopKeys.AddExStartStopKey('{$', '}', [], [fsItalic], DirectiveContext, ASMContext, 0, crDefault,
                                          clTeal, clWhite, False);

    StartStopKeys.AddExStartStopKey('{', '}', [], [fsItalic], ASMCommentContext, ASMContext, 0, crDefault,
                                          -1, clNavy, False);

    StartStopKeys.AddExStartStopKey('(*', '*)', [], [fsItalic], ASMCommentContext, ASMContext, 0, crDefault,
                                          -1, clGreen, False);

    StartStopKeys.AddExStartStopKey('asm', 'end', [woWholeWordsOnly], [fsUnderline], ASMContext, 0, 0, crDefault,
                                          -1, clPurple, False);

    { Create the THighlightInfo we use }
    CreateHighlight(fKeywordsOOP, [fsBold], -1, -1);
    CreateHighlight(fPreviousKeywordsOOP, [fsBold], -1, -1);
    fPreviousKeywordsOOP.OnChange:= nil;

    CreateHighlight(fKeywordsASM, [fsBold], clYellow, -1);
    CreateHighlight(fPreviousKeywordsASM, [fsBold], clYellow, -1);
    fPreviousKeywordsASM.OnChange:= nil;

    CreateHighlight(fASMHighlight, [], -1, clPurple);
    CreateHighlight(fStringsHilit, [fsItalic], clSilver, clBlack);
    CreateHighlight(fCommentSingleLine, [fsItalic], -1, clBlue);
    CreateHighlight(fDirectives, [fsItalic], clTeal, clWhite);
    CreateHighlight(fCommentBrace, [fsItalic], -1, clNavy);
    CreateHighlight(fCommentParAst, [fsItalic], -1, clGreen);
    CreateHighlight(fNumbers, [], -1, clRed);

    { Create internal TNumberHighlighter }
    fNumHilit:= TNumberHighlighter.Create(Self);
    with fNumHilit do
      begin
        Delimiters:= OOPDelimiters;
        ForeGround:= clRed;
        ContextNum:= NumberContext;
        OnQueryHilit:= GetNumberConfirmation;  { highlight numbers in asm and base parts as well }
      end;
  end;


destructor TOOPHighlighter.Destroy;
begin
  fAsmHighlight.Free;
  fStringsHilit.Free;
  fCommentBrace.Free;
  fCommentParAst.Free;
  fCommentSingleLine.Free;
  fDirectives.Free;
  fKeywordsASM.Free;
  fKeywordsOOP.Free;
  fPreviousKeywordsASM.Free;
  fPreviousKeywordsOOP.Free;
  fNumbers.Free;
  fKeywords.Free;
  fNumHilit.Free;
  inherited Destroy
end;

procedure TOOPHighlighter.Loaded;
var i: Integer; kinfo: TKeywordInfo; klist: TExtKeywordList;
begin
  fPreviousKeywordsOOP.Assign(fKeywordsOOP);
  fPreviousKeywordsASM.Assign(fKeywordsASM);

  { update internal keyword and start-stop key lists }

  { update OOP keywords, only those that were not streamed in
    for those having been read from stream, turn of the flag that says so. }
  klist:= inherited Keywords;
  kinfo.Options:= [woWholeWordsOnly];
  kinfo.Style:= fKeywordsOOP.Style;
  if fKeywordsOOP.AltFont then TPlusFontStyles(kinfo.Style):= TPlusFontStyles(kinfo.Style) + [TPlusFontStyle(fsAltFont)];
  kinfo.ContextNumber:= OOPKeywordContext;
  kinfo.Cursor:= crDefault;
  kinfo.Backgnd:= fKeywordsOOP.Background;
  kinfo.Foregnd:= fKeywordsOOP.Foreground;

  for i:= 0 to OOPKeywordCount-1 do
          with pExtKeyInfoLen(klist.Objects[i])^.BasicInfo, BasicPart do
            if Byte(Style) and $80 = 0 then BasicPart:= kinfo
                                       else Byte(Style):= Byte(Style) and $7f;

  { update ASM keywords, only those that were not streamed in }
  kinfo.Style:= fKeywordsASM.Style;
  if fKeywordsASM.AltFont then TPlusFontStyles(kinfo.Style):= TPlusFontStyles(kinfo.Style) + [TPlusFontStyle(fsAltFont)];
  kinfo.ContextNumber:= ASMKeywordContext;
  kinfo.Backgnd:= fKeywordsASM.Background;
  kinfo.Foregnd:= fKeywordsASM.Foreground;

  for i:= OOPKeywordCount to klist.Count-1 do
          with pExtKeyInfoLen(klist.Objects[i])^.BasicInfo, BasicPart do
            if Byte(Style) and $80 = 0 then BasicPart:= kinfo
                                       else Byte(Style):= Byte(Style) and $7f;

  fApplyHChange:= True;
  { update start-stop keys through the HighlightChange handler }
  HighlightChange(fASMHighlight);
  HighlightChange(fStringsHilit);
  HighlightChange(fCommentSingleLine);
  HighlightChange(fCommentBrace);
  HighlightChange(fCommentParAst);
  HighlightChange(fNumbers);
  HighlightChange(fDirectives);

  fApplyHChange:= False;
  inherited Loaded;    { note: this will also call ReApplyKeys }
end;

procedure TOOPHighlighter.setKeywordsOOP(h: THighlightInfo);
begin
  fKeywordsOOP.Assign(h)
end;

procedure TOOPHighlighter.setKeywordsASM(h: THighlightInfo);
begin
  fKeywordsASM.Assign(h)
end;

procedure TOOPHighlighter.setASMHighlight(h: THighlightInfo);
begin
  fASMHighlight.Assign(h)
end;

procedure TOOPHighlighter.setStringsHilit(h: THighlightInfo);
begin
  fStringsHilit.Assign(h)
end;

procedure TOOPHighlighter.setCommentSingleLine(h: THighlightInfo);
begin
  fCommentSingleLine.Assign(h)
end;

procedure TOOPHighlighter.setCommentBrace(h: THighlightInfo);
begin
  fCommentBrace.Assign(h)
end;

procedure TOOPHighlighter.setCommentParAst(h: THighlightInfo);
begin
  fCommentParAst.Assign(h)
end;

procedure TOOPHighlighter.setDirectives(h: THighlightInfo);
begin
  fDirectives.Assign(h)
end;

procedure TOOPHighlighter.setNumbers(h: THighlightInfo);
begin
  fNumbers.Assign(h)
end;

procedure TOOPKeywordList.DefineProperties(Filer: TFiler);
{ we stream only the format of keywords, and then only those
  that differ from the base setting of the parent TOOPHighlighter,
  that's the reason for this override }
begin
  Filer.DefineProperty('Formats', TOOPHighlighter(ParentHighlighter).ReadOldFormats, WriteData, False)
end;


procedure TOOPKeywordList.WriteData(Writer: TWriter);
begin
  { nothing to do }
end;


procedure TOOPHighlighter.HighlightChange(Sender: TObject);
  { event handler for internal THighlightInfo modifications }

  procedure setkeyword(var kw: TExtKeyInfoLen; oldHighlight, newHighlight: THighlightInfo);
    begin    { set keyword formatting info according to values of newHighlight, only
               for attributes that differ from oldHighlight }
      kw.BasicInfo.BasicPart.Style:= kw.BasicInfo.BasicPart.Style + (newHighlight.Style-oldHighlight.Style)
                                    - (oldHighlight.Style-newHighlight.Style);

      if oldHighlight.AltFont<>newHighlight.AltFont then
        if newHighlight.AltFont then Include(kw.BasicInfo.BasicPart.Style, TExtFontStyle(fsAltFont))
                                else Exclude(kw.BasicInfo.BasicPart.Style, TExtFontStyle(fsAltFont));

      if oldHighlight.Background<>newHighlight.Background then
                              kw.BasicInfo.BasicPart.Backgnd:= newHighlight.Background;

      if oldHighlight.Foreground<>newHighlight.Foreground then
                              kw.BasicInfo.BasicPart.Foregnd:= newHighlight.Foreground
    end;

  procedure setssinfo(var ss: TExtStartStopInfo; aHighlight: THighlightInfo);
    begin   { set ss formatting info according to values of aHighlight }
      ss.BasicSSInfo.Attributes.Style:= aHighlight.Style;
      if aHighlight.AltFont then Include(ss.BasicSSInfo.Attributes.Style, TExtFontStyle(fsAltFont));
      ss.BasicSSInfo.Attributes.Backgnd:= aHighlight.Background;
      ss.BasicSSInfo.Attributes.Foregnd:= aHighlight.Foreground
    end;

var i, ssupdate: Integer; klist: TExtKeywordList;
begin    { HighlightChange }
  if (csLoading in ComponentState) and (not fApplyHChange) then Exit;

   if Sender=fKeywordsOOP then
    begin
      if csDesigning in ComponentState then
        if not ConfirmKeywords(fKeywordsOOP, fPreviousKeywordsOOP, 'OOP') then Exit;

      { update the OOP keywords highlight attributes }
      klist:= inherited Keywords;
      for i:= 0 to OOPKeywordCount-1 do
          setkeyword(pExtKeyInfoLen(klist.Objects[i])^, fPreviousKeywordsOOP, fKeywordsOOP);
      fPreviousKeywordsOOP.Assign(fKeywordsOOP)
    end;

  if Sender=fKeywordsASM then
    begin
      if csDesigning in ComponentState then
        if not ConfirmKeywords(fKeywordsASM, fPreviousKeywordsASM, 'ASM') then Exit;

      { update the ASM keywords highlight attributes }
      klist:= inherited Keywords;
      for i:= OOPKeywordCount to klist.Count-1 do
          setkeyword(pExtKeyInfoLen(klist.Objects[i])^, fPreviousKeywordsASM, fKeywordsASM);
      fPreviousKeywordsASM.Assign(fKeywordsASM)
    end;

  if Sender=fASMHighlight then
    setssinfo(pExStartStopInfo(StartStopKeys.Pointers[AsmSSIndex])^, fASMHighlight);

  { determine which start-stop key needs to be updated }
  ssupdate:= -1;

  if Sender=fCommentBrace then ssupdate:= 4;
  if Sender=fCommentParAst then ssupdate:= 5;
  if Sender=fCommentSingleLine then ssupdate:= 2;
  if Sender=fDirectives then ssupdate:= 3;
  if Sender=fStringsHilit then ssupdate:= 1;

  if ssupdate>=0 then
    begin
      setssinfo(pExStartStopInfo(StartStopKeys.Pointers[ssupdate])^, THighlightInfo(Sender));
      setssinfo(pExStartStopInfo(StartStopKeys.Pointers[ssupdate+5])^, THighlightInfo(Sender))
    end;

  if Sender=fNumbers then
    with fNumHilit do
      begin
        Background:= fNumbers.Background;
        Foreground:= fNumbers.Foreground;
        Style:= fNumbers.Style;
        AltFont:= fNumbers.AltFont
      end;

  if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
end;

procedure TOOPHighlighter.GetKeywordConfirmation(Sender: TObject; StartKey,
  StopKey: TPlusNavigator; KIndex: Integer; var Accept: Boolean);
begin
  case KIndex of
    AsmKIndex:   Accept:= StartKey.ParOffset=StartKey.pDynAttr^.DynOffset;
                 { asm keyword, highlight only if found at start of asm section }

    PropertyKeywordsIndex..OOPKeywordCount-1:
                 Accept:= StartKey.pDynAttr^.DynStyle and $c0 = $c0
                 { read, write and default keywords highlighted only in property sections }
  end;
end;

procedure TOOPHighlighter.GetNumberConfirmation(Sender: TObject; Scope: SmallInt; var DoHilit: Boolean);
begin
  DoHilit:= (Scope=0) or (Scope=ASMContext)
end;


procedure TOOPHighlighter.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('ModFormats', ReadNewFormats, WriteFormats, True)
end;

procedure TOOPHighlighter.ReadNewFormats(Reader: TReader);
var klist: TExtKeywordList; kindex: Integer; k: TKeywordInfo;
begin
  klist:= inherited Keywords;
  Reader.ReadListBegin;
  while not Reader.EndOfList do
    begin
      kindex:= Reader.ReadInteger;
      with k do
        begin
          Byte(Options):= Reader.ReadInteger;
          Byte(Style):= Reader.ReadInteger or $80; { a flag to mark as loaded from stream }
          ContextNumber:= Reader.ReadInteger;
          Cursor:= Reader.ReadInteger;
          Backgnd:= Reader.ReadInteger;
          Foregnd:= Reader.ReadInteger
        end;
      if kindex<klist.Count then pExtKeyInfoLen(klist.Objects[kindex])^.BasicInfo.BasicPart:= k;
    end;
  Reader.ReadListEnd;
end;

procedure TOOPHighlighter.ReadOldFormats(Reader: TReader);
{ reads pre-4.1 streams }
var k: TKeywordInfo; kindex: Integer; klist: TExtKeywordList;
begin
  klist:= inherited Keywords;
  Reader.ReadListBegin;
  kindex:= 0;
  while not Reader.EndOfList do
    begin
      with k do
        begin
          Byte(Options):= Reader.ReadInteger;
          Byte(Style):= Reader.ReadInteger or $80; { flag it as loaded from stream }
          ContextNumber:= Reader.ReadInteger;
          Cursor:= Reader.ReadInteger;
          Backgnd:= Reader.ReadInteger;
          Foregnd:= Reader.ReadInteger
        end;
      if kindex<klist.Count then pExtKeyInfoLen(klist.Objects[kindex])^.BasicInfo.BasicPart:= k;
      Inc(kindex)
    end;
  Reader.ReadListEnd;
end;

procedure TOOPHighlighter.WriteFormats(Writer: TWriter);
  function samehinfo(const k: TKeywordInfo; hilit: THighlightInfo): Boolean;
    var st: TExtFontStyles;
    begin
      st:= hilit.Style;
      if hilit.AltFont then TPlusFontStyles(st):= TPlusFontStyles(st) + [TPlusFontStyle(fsAltFont)];
      Result:= (k.Backgnd=hilit.Background) and (k.Foregnd=hilit.Foreground) and (k.Style=st)
    end;

var i: Integer; klist: TExtKeywordList; h: THighlightInfo;

type pKeyInfo = ^TKeywordInfo;
begin
  Writer.WriteListBegin;
  klist:= inherited Keywords;
  for i:= 0 to klist.Count-1 do
    with pExtKeyInfoLen(klist.Objects[i])^.BasicInfo do
      begin
        if i<OOPKeywordCount then h:= KeywordsOOP
                             else h:= KeywordsASM;
        if not samehinfo(BasicPart, h) then
          begin
            Writer.WriteInteger(i);
            Writer.WriteInteger(Byte(BasicPart.Options));
            Writer.WriteInteger(Byte(BasicPart.Style));
            Writer.WriteInteger(BasicPart.ContextNumber);
            Writer.WriteInteger(BasicPart.Cursor);
            Writer.WriteInteger(BasicPart.Backgnd);
            Writer.WriteInteger(BasicPart.Foregnd)
          end
      end;
  Writer.WriteListEnd
end;

type TPNumHilit = class(TNumberHighlighter);  { for accessing its ApplyKeywordsList method }

procedure TOOPHighlighter.ApplyKeywordsList(Start, Stop: TPlusNavigator; BaseIndex: Integer);
begin
  inherited ApplyKeywordsList(Start, Stop, BaseIndex);
  TPNumHilit(fNumHilit).ApplyKeywordsList(Start, Stop, BaseIndex)
end;

end.
