unit PlusToFormat;
{  Electro-Concept Mauricie, 1999 }
{ Implements TPlusToRTF and TPlusToHtml components, used to get the content of a TPlusMemo
  in RTF or HTML format.

  See PlusToFormat.hlp for information on how to use them. }

interface

uses
  Classes, PlusMemo;

type
  TPlusToFormattedText = class(TComponent)
    private
      fDynFormat   : Boolean;
      fPlusMemo    : TPlusMemo;
      fOnProgress  : TNotifyEvent;
      function getSelection: string;
      function getText: string;
    protected
      function ClipboardFormat: Integer; virtual;
      procedure getFormattedText(Nav1, Nav2: TPlusNavigator; Stream: TStream; sList: TStringList); virtual; abstract;
      procedure DoDynHilit(Nav1, Nav2: TPlusNavigator);
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    public
      constructor Create(AOwner: TComponent); override;
      procedure CopyToClipboard; virtual;
      procedure SaveToStream(Stream: TStream);
      procedure SaveToFile(FName: string);
      property Text: string read getText;
      property Selection: string read getSelection;
    published
      property DynFormat: Boolean read fDynFormat write fDynFormat;
      property PlusMemo: TPlusMemo read fPlusMemo write fPlusMemo;
      property OnProgress: TNotifyEvent read fOnProgress write fOnProgress;
    end;

  TPlusToRTF = class(TPlusToFormattedText)
    private
      fRTFLeading  : Integer;
    protected
      function ClipboardFormat: Integer; override;
      procedure getFormattedText(Nav1, Nav2: TPlusNavigator; Stream: TStream; sList: TStringList); override;
    public
      constructor Create(AOwner: TComponent); override;
    published
      property RTFLeading: Integer read fRTFLeading write fRTFLeading;
    end;

  TGetSizeEvent = procedure (Sender: TObject; FontSize: Integer; var HtmlSize: string) of object;

  TPlusToHTML = class(TPlusToFormattedText)
    private
      fBodyTag: Boolean;
      fHeadTag: Boolean;
      fTitle: string;
      fBaseFont: Boolean;
      FOnGetSize: TGetSizeEvent;
      fDocType: string;
      function SizeToString(Size: Integer): string;
    protected
      function ClipboardFormat: Integer; override;
      procedure getFormattedText(Nav1, Nav2: TPlusNavigator; Stream: TStream; sList: TStringList); override;
    public
      constructor Create(AOwner: TComponent); override;
    published
      property BodyTag: Boolean read fBodyTag write fBodyTag;
      property HeadTag: Boolean read fHeadTag write fHeadTag;
      property DocType: string read fDocType write fDocType;
      property Title: string read fTitle write fTitle;
      property BaseFont: Boolean read fBaseFont write fBaseFont;
      property OnGetSize: TGetSizeEvent read fOnGetSize write fOnGetSize;
    end;

  procedure Register;

implementation

uses SysUtils, Windows, Graphics, Clipbrd;

var RTFClipboardFormat, HtmlClipboardFormat: Integer;   { set to zero at init time }

procedure Register;
begin
  RegisterComponents('ECM', [TPlusToRTF, TPlusToHtml]);
end;

type TProtPMemo = class(TPlusMemo);   { to have access to protected members }

{ TPlusToFormattedText }

function TPlusToFormattedText.ClipboardFormat: Integer;
begin
  Result:= CF_TEXT
end;

procedure TPlusToFormattedText.CopyToClipboard;
var
  copytext: string;
  Data: THandle;
  DataPtr: Pointer;
  cformat: Integer;
begin
  if (PlusMemo=nil) or (PlusMemo.SelLength=0) then Exit;
  cformat:= ClipboardFormat;
  copytext:= Selection;
  Data := GlobalAlloc(GMEM_MOVEABLE+GMEM_DDESHARE, Length(copytext)+1);
  try
    DataPtr := GlobalLock(Data);
    try
      Move(copytext[1], DataPtr^, Length(copytext)+1);
      Clipboard.SetAsHandle(cformat, Data);
    finally
      GlobalUnlock(Data);
    end;
  except
    GlobalFree(Data);
    raise;
  end;
end;

constructor TPlusToFormattedText.Create(AOwner: TComponent);
begin
  inherited;
  fDynFormat:= True
end;

procedure TPlusToFormattedText.DoDynHilit(Nav1, Nav2: TPlusNavigator);
var i, dummy, dummy1: Integer; dummyf: THandle;
begin
  if DynFormat then
    with TProtPMemo(Nav1.fPMemo) do
      begin
        ParseStartStopNow(Nav2.ParNumber);
        for i:= Nav1.ParNumber to Nav2.ParNumber do
          with fParagraphs.ParPointers[i]^ do
            if not KeywordsDone then
              begin
                ReformatPar(0, False, 0, nil, i, dummy, dummy1, True, dummyf);
                Formatted:= False
              end
      end;

end;

function TPlusToFormattedText.getSelection: string;
var tmpstrings: TStringList;
begin
  if (PlusMemo=nil) or (PlusMemo.SelLength=0) then Result:= ''
  else
    begin
      tmpstrings:= TStringList.Create;
      with TProtPMemo(PlusMemo) do
          getFormattedText(fSelStart, fSelStop, nil, tmpstrings);
      Result:= tmpstrings.Text;
      tmpstrings.Free
    end
end;

function TPlusToFormattedText.getText: string;
var tmpstrings: TStringList; start, stop: TPlusNavigator;
begin
  if (PlusMemo=nil) or (PlusMemo.CharCount=0) then Result:= ''
  else
    begin
      tmpstrings:= TStringList.Create;
      start:= TPlusNavigator.Create(PlusMemo);
      stop:= TPlusNavigator.Create(PlusMemo);
      stop.Pos:= PlusMemo.CharCount;
      getFormattedText(start, stop, nil, tmpstrings);
      Result:= tmpstrings.Text;
      tmpstrings.Free;
      start.Free;
      stop.Free
    end
end;

procedure TPlusToFormattedText.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if (Operation=opRemove) and (AComponent=PlusMemo) then
      fPlusMemo:= nil;
  inherited;
end;

procedure TPlusToFormattedText.SaveToFile(FName: string);
var f: TFileStream;
begin
  if PlusMemo=nil then Exit;
  f:= TFileStream.Create(FName, fmCreate);
  try
    SaveToStream(f);
  finally
    f.Free
  end
end;

procedure TPlusToFormattedText.SaveToStream(Stream: TStream);
var start, stop: TPlusNavigator;
begin
  if PlusMemo=nil then Exit;
  start:= TPlusNavigator.Create(PlusMemo);
  stop:= TPlusNavigator.Create(PlusMemo);
  stop.Pos:= PlusMemo.CharCount;
  GetFormattedText(start, stop, Stream, nil)
end;


procedure XORStyleCode(var Style: TExtFontStyles; Code: Char);
var element: TExtFontStyle;  isformatchar: Boolean;
begin
  isformatchar:= True;
  element:= fsBold;  { to avoid a warning }
  case Code of
    ctrlUnderline: element:= fsUnderline;
    ctrlItalic   : element:= fsItalic;
    ctrlBold     : element:= fsBold;
    ctrlHighlight: element:= TExtFontStyle(fsHighlight);
    ctrlAltFont  : element:= TExtFontStyle(fsAltFont)
    else isformatchar:= False
    end;

  if isformatchar then
    if element in Style then Exclude(Style, element)
                        else Include(Style, element)
end;



{ TPlusToRTF }

function TPlusToRTF.ClipboardFormat: Integer;
begin
  if RTFClipboardFormat=0 then
      RTFClipboardFormat:= RegisterClipboardFormat('Rich Text Format');
  Result:= RTFClipboardFormat
end;

constructor TPlusToRTF.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fRTFLeading:= -1;
end;

procedure TPlusToRTF.getFormattedText(Nav1, Nav2: TPlusNavigator; Stream: TStream; sList: TStringList);
const SpecialChars = ['{', '}', '\'];
      crlf: string[2] = #13#10;

var smemo: TProtPMemo;
    colortable: TList;

  procedure Output(const s: string);
    begin
      if Stream<>nil then
        begin
          if s<>'' then Stream.WriteBuffer(s[1], Length(s));
          Stream.WriteBuffer(crlf, 2)
        end;
      if sList<>nil then sList.Add(s)
    end;

  function RTFColor(c: TColor): string;
    var crgb: Longint;
    begin
      crgb:= ColorToRGB(c);
      Result:= '\red' + IntToStr(crgb and $0FF) +
               '\green' + IntToStr((crgb and $0FF00) shr 8) +
               '\blue' + IntToStr((crgb and $FF0000) shr 16) +
               ';'
    end;


  procedure buildcolortable;  { add to color table the values needed for dyn parts }
    begin
      with smemo.ftmpnav1 do
        begin
          Assign(Nav1);
          repeat
            RightOfDyn;
            with DynAttr do
              if (DynStyle and $80 <>0) and ((Backgnd<>0) or (Foregnd<>0)) then
                  begin
                    if colortable.IndexOf(Pointer(Backgnd))<0 then colortable.Add(Pointer(Backgnd));
                    if colortable.IndexOf(Pointer(Foregnd))<0 then colortable.Add(Pointer(Foregnd))
                  end
          until not ForwardToDyn(Nav2.Pos)
        end
    end;


  function RTFFormat(Style: TExtFontStyles; const dyn: DynInfoRec): string;
    var sb: Byte; dyncolor: Boolean;
    begin
      Result:= '\plain';
      dyncolor:= False;

      if (dyn.DynStyle and $80<>0) and DynFormat then
        begin
          sb:= dyn.DynStyle and $3f;
          Style:= TExtFontStyles(sb);
          dyncolor:= (dyn.Foregnd<>0) or (dyn.Backgnd<>0)
        end;
      if TPlusFontStyle(fsAltFont) in TPlusFontStyles(Style) then
          Result:= Result + '\f1\fs' + IntToStr(Abs(smemo.AltFont.Size)*2)
      else
          Result:= Result + '\f0\fs' + IntToStr(Abs(smemo.Font.Size)*2);
      { N.B: The font size in RTF for TrueType is the double as normal font }

      if fsBold in Style then Result:= Result + '\b';
      if fsItalic in Style then Result:= Result + '\i';
      if fsUnderline in Style then Result:= Result + '\ul';
      if not dyncolor then
          if TPlusFontStyle(fsHighlight) in TPlusFontStyles(Style) then Result:= Result + '\cf1\cb3'
                                                                   else Result:= Result + '\cf0\cb2'
      else
          Result:= Result + '\cf' + IntToStr(colortable.IndexOf(Pointer(dyn.Foregnd))) +
                            '\cb'+ IntToStr(colortable.IndexOf(Pointer(dyn.Backgnd)));

      Result:= Result + ' '
    end;

var i, lastnotify: Longint;
    j, jlim,
    lastwritten,
    lastchar, tlen: PlusCardinal;
    s, olds, st, s1: string;
    parp          : pParInfo;
    t             : PChar;
    CurrentStyle  : TExtFontStyles;
    currentdyn    : DynInfoRec;
    currdynnb     : Integer;
    formatmodified: Boolean;
    fs            : TFontStyles;

begin
  DoDynHilit(Nav1, Nav2);
  smemo:= TProtPMemo(nav1.fPMemo);
  lastnotify:= GetTickCount;
  S:= '{\rtf1\ansi\deff0\deftab720'+          { head of an rtf an open { }
      '{\fonttbl{\f0\fnil '+smemo.Font.Name+';}' +  { First font 0 Index }
      '{\f1\fnil '+smemo.AltFont.Name+';}}';        { uses another font }
  Output(s);

  colortable:= TList.Create;
  { Write the color table: 0 is for normal text, 1 for highlighted text,
                           2 normal background,  3 for highlighted backg. }
  colortable.Add(Pointer(smemo.Font.Color));
  colortable.Add(Pointer(smemo.HighlightColor));
  colortable.Add(Pointer(smemo.Color));
  colortable.Add(Pointer(smemo.HighlightBackgnd));
  Output('{\colortbl');
  if DynFormat then buildcolortable;

  s:= '';
  for i:= 0 to colortable.Count-1 do s:= s + RTFColor(TColor(colortable[i]));
  s:= s+ '}';

  { Alignment of the text, we make general because TPlusMemo
    does not support paragraph format }
  if smemo.Justified then s:= s + '\pard\qj'
  else case smemo.Alignment of
        taLeftJustify : s:= s+'\pard\ql';
        taRightJustify: s:= s+'\pard\qr'
       else s:= s+'\pard\qc'   { center }
       end;

  { leading, written in twips }
  if RTFLeading>0 then st:= s + '\sl'+IntToStr(RTFLeading);
  if RTFLeading<0 then st:= s + '\sl'+IntToStr(Abs(smemo.Font.Size)*20*RTFLeading);

  { write initial formatting info }
  fs:= smemo.Font.Style;
  CurrentStyle:= Nav1.Style;
  currentdyn:= Nav1.DynAttr;
  s:= RTFFormat(CurrentStyle, currentdyn);
  Output(st+s);
  st:= '';

  { go with the text content }
  for i:= Nav1.ParNumber to Nav2.ParNumber do
    begin
      parp:= smemo.fParagraphs.ParPointers[i];
      t:= parp^.Text;
      if i=Nav1.fParNb then
        begin
          currdynnb:= Nav1.DynNb;
          j:= Nav1.ParOffset;
          st:= ''
        end
      else
        begin
          j:= 0;
          currdynnb:= 0;
          st:= '\par '
        end;

      if DynFormat and (parp^.DynCount>currdynnb) then jlim:= parp^.DynBuffer^[currdynnb].DynOffset
                                                  else jlim:= High(jlim);
      if i=Nav2.fParNb then tlen:= Nav2.ParOffset
                       else tlen:= parp^.Length;

      lastwritten:= j; lastchar:= j;

      while j<tlen do
        begin
          formatmodified:= False;
          if t[j] in SpecialChars then
            begin
              SetLength(s1, j-lastwritten);
              if j>lastwritten then Move(t[lastwritten], s1[1], j-lastwritten);
              st:= st + s1 + '\' + t[j];
              Inc(j);
              lastwritten:= j;
              lastchar:= j
            end

          else
            if smemo.StaticFormat and (t[j]<#26) and (t[j] in CtrlCodesSet) then
              begin
                formatmodified:= True;
                repeat
                  XORStyleCode(CurrentStyle, t[j]);
                  Inc(j)
                until not (t[j] in CtrlCodesSet)
              end
            else
              begin
                Inc(j);
                lastchar:= j
              end;

          if j>=jlim then
            begin
              Inc(currdynnb);
              while (currdynnb<parp^.DynCount) and (parp^.DynBuffer^[currdynnb].DynOffset<=j) do Inc(currdynnb);
              currentdyn:= parp^.DynBuffer^[currdynnb-1];
              if currdynnb<parp^.DynCount then jlim:= parp^.DynBuffer^[currdynnb].DynOffset
                                          else jlim:= High(jlim);
              formatmodified:= True
            end;

          if formatmodified then
            begin
              SetLength(s1, lastchar-lastwritten);
              if lastchar>lastwritten then Move(t[lastwritten], s1[1], lastchar-lastwritten);
              st:= st + s1;
              olds:= s;
              s:= RTFFormat(CurrentStyle, currentdyn);
              if s<>olds then st:= st + s;
              lastwritten:= j
            end
        end;  { while j<tlen }

      SetLength(s1, j-lastwritten);
      if j>lastwritten then Move(t[lastwritten], s1[1], j-lastwritten);
      st:= st + s1;
      Output(st);
      if Assigned(fOnProgress) then
        if Longint(GetTickCount)-lastnotify>ProgressInterval then
          begin
            fOnProgress(Self);
            lastnotify:= GetTickCount
          end
    end;  { i loop over paragraphs }

  { write RTF end part }
  Output('\par }'#0);

  if Assigned(fOnProgress) then fOnProgress(Self);
  colortable.Free
end;


{ TPlusToHTML }

const BRTag = '<BR>';
      NbspTag = '&nbsp;';


type SpecCharRec = record c: Char; Tag: string end;

const SpecCharCount = 5;
      SpecChars: array[0..SpecCharCount-1] of SpecCharRec =
        ((c: '>'; Tag: '&gt;'),
         (c: '<'; Tag: '&lt;'),
         (c: '&'; Tag: '&amp;'),
         (c: '"'; Tag: '&quot;'),
         (c: #9;  Tag: '&#9;'));

var SpecialChars: TSetOfChar;   { initialized at startup from elements found SpecChars }


function ColorToString(color: TColor): string;
begin
  if color=clNone then Result:= ''
  else
    begin
      Result:= IntToHex(ColorToRGB(color), 6);
      Result:= '"#' + Copy(Result, 5,2) + Copy(Result, 3, 2) + Copy(Result, 1, 2)+'"'
    end
end;


function TagFromChar(car: Char): string;
var i: Integer;
begin
  Result:= '';
  for i:= 0 to SpecCharCount-1 do
    if SpecChars[i].c=car then
      begin
        Result:= SpecChars[i].Tag;
        Exit
      end
end;

function GetEndTag(element: TExtFontStyle): string;
begin
  case Byte(element) of
    Byte(fsBold): Result:= '</B>';
    Byte(fsItalic): Result:= '</I>';
    Byte(fsUnderline): Result:= '</U>';
    Byte(fsStrikeOut): Result:= '</S>';
    Byte(fsHighlight), Byte(fsAltFont): Result:= '</FONT>'
    else Result:= ''
  end
end;

function GetStartTag(element: TExtFontStyle; fontcolor: TColor; fontname, fontsize, fontattr: string): string;
begin
  case Byte(element) of
    Byte(fsBold): Result:= '<B>';
    Byte(fsItalic): Result:= '<I>';
    Byte(fsUnderline): Result:= '<U>';
    Byte(fsStrikeOut): Result:= '<S>';
    Byte(fsAltFont): begin
                 Result:= '<FONT ';
                 if (fontname<>'') or (fontsize<>'') or (fontattr<>'') then
                     if fontattr='' then
                       begin
                         if fontname<>'' then Result:= Result + 'FACE="' + fontname+'"';
                         if fontsize<>'' then Result:= Result + 'SIZE="' + fontsize+'"'
                       end
                     else
                       Result:= Result + fontattr;
                 Result:= Result + '>'
               end;

    Byte(fsHighlight): Result:= '<FONT COLOR=' + ColorToString(fontcolor) + '>'
    else Result:= ''
  end
end;


function TPlusToHTML.ClipboardFormat: Integer;
begin
  if HtmlClipboardFormat=0 then
    HtmlClipboardFormat:= RegisterClipboardFormat('HTML Format');
  Result:= HtmlClipboardFormat
end;

constructor TPlusToHTML.Create(AOwner: TComponent);
begin
  inherited;
  DocType:= '-//W3C//DTD HTML 3.2 Final//EN';
  HeadTag:= True;
  BodyTag:= True;
  BaseFont:= True;
  Title:= 'PlusToHtml view';
end;

procedure TPlusToHTML.getFormattedText(Nav1, Nav2: TPlusNavigator; Stream: TStream; sList: TStringList);
var stylestack : array[0..Ord(High(TExtFontStyle))] of TExtFontStyles;
    stylelevel : Integer;
    accumulator: PChar;
    acclen     : Integer;
    smemo      : TProtPMemo;

  procedure AccumulateBuf(const Buf; Len: Integer);
    var newacc: PChar;
    begin
      if accumulator=nil then
        begin
          accumulator:= StrAlloc(Len);
          acclen:= 0
        end;

      if Integer(StrBufSize(accumulator))<acclen+Len then
        begin
          newacc:= StrAlloc(acclen+Len);
          Move(accumulator^, newacc^, acclen);
          StrDispose(accumulator);
          accumulator:= newacc
        end;
      Move(Buf, accumulator[acclen], Len);
      Inc(acclen, Len)
    end;

  procedure AccumulateString(const s: string);
    begin
      AccumulateBuf(s[1], Length(s))
    end;

  procedure StreamOut(s: string);
    const crlf: string[2] = #13#10;
    var oldlen: Integer;
    begin
      if Stream<>nil then
        begin
          if accumulator<>nil then Stream.WriteBuffer(accumulator^, acclen);
          if s<>'' then Stream.WriteBuffer(s[1], Length(s));
          Stream.WriteBuffer(crlf[1], 2)
        end;
      if sList<>nil then
        begin
          if (accumulator<>nil) and (acclen>0) then
            begin
              oldlen:= Length(s);
              SetLength(s, Length(s)+acclen);
              Move(s[1], s[acclen+1], oldlen);
              Move(accumulator^, s[1], acclen)
            end;
          sList.Add(s)
        end;
      acclen:= 0
    end;

  procedure HTMLFormat(Style: TExtFontStyles; const dyn: DynInfoRec; var currentcolor: TColor);
    var sb: Byte;
        dyncolor: Boolean;
        laststyle, turnoff, turnon: TExtFontStyles;
        i: TExtFontStyle;
        slevel: Integer;
        newfontname, newsize: string;
        newcolor: TColor;
    begin
      dyncolor:= False;
      if (dyn.DynStyle and $80<>0) and DynFormat then
        begin
          sb:= dyn.DynStyle and $3f;
          Style:= TExtFontStyles(sb);
          dyncolor:= (dyn.Foregnd<>-1) and (dyn.Foregnd<>clNone)
        end;

      if dyncolor then
        begin
          newcolor:= dyn.Foregnd;
          Include(Style, TExtFontStyle(fsHighlight))
        end
      else
        if TExtFontStyle(fsHighlight) in Style then newcolor:= smemo.HighlightColor
                                               else newcolor:= clNone;

      { find elements we have to turn off }
      if stylelevel>=0 then laststyle:= stylestack[stylelevel]
                       else laststyle:= [];
      turnoff:= laststyle-Style;

      { find stack level to unwind }
      slevel:= stylelevel;
      while laststyle*turnoff<>[] do
        begin
          Dec(slevel);
          if slevel>=0 then laststyle:= stylestack[slevel]
                       else laststyle:= []
        end;

      { output end tags }
      if slevel<stylelevel then turnoff:= turnoff + (stylestack[stylelevel]-laststyle);
      for i:= High(TExtFontStyle) downto Low(TExtFontStyle) do
        if i in turnoff then
            AccumulateString{StreamOut}(GetEndTag(i));

      { output attribute tags, put them on the stack }
      if TExtFontStyle(fsAltFont) in Style then
        begin
          newfontname:= smemo.AltFont.Name;
          newsize:= SizeToString(smemo.AltFont.Size)
        end
      else
        begin
          newfontname:= '';
          newsize:= ''
        end;

      turnon:= Style-laststyle;
      for i:= Low(TExtFontStyle) to High(TExtFontStyle) do
        if i in turnon then
          begin
            Include(laststyle, i);
            Inc(slevel);
            stylestack[slevel]:= laststyle;
            AccumulateString(GetStartTag(i, newcolor, newfontname, newsize, ''))
          end;

      stylelevel:= slevel;
      currentcolor:= newcolor
    end;

var i              : Longint;
    j, jlim, navlim,
    lastwritten,
    lastchar, tlen: PlusCardinal;
    lastnotify    : Longint;
    parp          : pParInfo;
    t             : PChar;
    CurrentStyle  : TExtFontStyles;
    currentdyn    : DynInfoRec;
    currdynnb     : Integer;
    runningcolor: TColor;
    formatmodified: Boolean;
    fs            : TFontStyles;
    dummy, dummy1 : Integer;
    dummyf        : THandle;
    full          : Boolean;
    s             : string;
    dynnblim      : Integer;

const AlignmentToStr: array[TAlignment] of string = ('LEFT', 'RIGHT', 'CENTER');

begin
  accumulator:= nil;
  DoDynHilit(Nav1, Nav2);
  smemo:= TProtPMemo(Nav1.fPMemo);
  lastnotify:= GetTickCount;
  stylelevel:= -1;
  full:= (Nav1.Pos=0) and (Nav2.Pos=smemo.CharCount);

  if full then
    begin
      if DocType<>'' then
        StreamOut('<!DOCTYPE HTML PUBLIC "' + DocType + '">');

      if HeadTag then
        begin
          StreamOut('<HTML>');
          StreamOut('<HEAD>')
        end;

      if Title<>'' then
        StreamOut('<TITLE>' + Title + '</TITLe>');

      if HeadTag then StreamOut('</HEAD>')
    end;

  if BodyTag then
    begin
      StreamOut('<BODY BGCOLOR=' + ColorToString(smemo.Color) +
                            ' TEXT=' + ColorToString(smemo.Font.Color) + '>');
      runningcolor:= smemo.Font.Color
    end
  else
    runningcolor:= clNone;

  if BaseFont then
    begin
      s:= 'FACE="' + smemo.Font.Name + '" SIZE=' + SizeToString(smemo.Font.Size);
       StreamOut('<BASEFONT '+s+'>')
    end;

  StreamOut('<DIV ALIGN='+AlignmentToStr[smemo.Alignment]+'>');

  fs:= Nav1.Style;
  CurrentStyle:= TExtFontStyles(fs);
  if DynFormat then currentdyn:= Nav1.DynAttr else currentdyn.DynStyle:= 0;
  runningcolor:= clNone;
  HTMLFormat(CurrentStyle, currentdyn, runningcolor);

  { go with the text content }
  for i:= Nav1.ParNumber to Nav2.ParNumber do
    begin
      parp:= smemo.fParagraphs.Pointers[i];
      t:= parp^.Text;

      if i=Nav1.fParNb then
        begin
          currdynnb:= Nav1.DynNb;
          j:= Nav1.fOffset
        end
      else
        begin
          currdynnb:= 0;
          j:= 0
        end;

      if i=Nav2.fParNb then
        begin
          tlen:= Nav2.fOffset;
          dynnblim:= Nav2.DynNb
        end
      else
        begin
          tlen:= parp^.Length;
          dynnblim:= parp^.DynCount
        end;

      if DynFormat and (currdynnb<dynnblim) then jlim:= parp^.DynBuffer^[currdynnb].DynOffset
                                            else jlim:= High(jlim);

      lastwritten:= j;
      lastchar:= j;

      if jlim<=j then
        begin
          Inc(currdynnb);
          while (currdynnb<dynnblim) and (parp^.DynBuffer^[currdynnb].DynOffset=0) do Inc(currdynnb);
          currentdyn:= parp^.DynBuffer^[currdynnb-1];
          if currdynnb<dynnblim then jlim:= parp^.DynBuffer^[currdynnb].DynOffset
                                else jlim:= High(jlim);
          HtmlFormat(CurrentStyle, currentdyn, runningcolor);
        end;

      while j<tlen do
        begin
          formatmodified:= False;
          if (t[j]<#26) and (t[j] in CtrlCodesSet) and smemo.StaticFormat then
            begin
              formatmodified:= True;
              repeat
                XORStyleCode(CurrentStyle, t[j]);
                Inc(j)
              until not (t[j] in CtrlCodesSet)
            end
          else
           begin
             if ((t[j]=' ') and (t[j+1]=' ')) or (t[j] in SpecialChars) then
               begin    { output a non breaking space }
                 AccumulateBuf(t[lastwritten], lastchar-lastwritten);
                 if t[j]=' ' then AccumulateString(NbspTag)
                             else AccumulateString(TagFromChar(t[j])); 
                 lastwritten:= j+1
               end;
             Inc(j);
             lastchar:= j
           end;

          if j>=jlim then
            begin
              Inc(currdynnb);
              while (currdynnb<dynnblim) and (parp^.DynBuffer^[currdynnb].DynOffset<=j) do Inc(currdynnb);
              currentdyn:= parp^.DynBuffer^[currdynnb-1];
              if currdynnb<dynnblim then jlim:= parp^.DynBuffer^[currdynnb].DynOffset
                                    else jlim:= High(jlim);
              formatmodified:= True
            end;

          if formatmodified then
            begin
              AccumulateBuf(t[lastwritten], lastchar-lastwritten);
              lastwritten:= j;
              lastchar:= j;
              HtmlFormat(CurrentStyle, currentdyn, runningcolor);
            end;
        end;  { while j<tlen }

      AccumulateBuf(t[lastwritten], j-lastwritten);
      StreamOut(BRTag);
      if Assigned(fOnProgress) then
        if Longint(GetTickCount)-lastnotify>ProgressInterval then
          begin
            fOnProgress(Self);
            lastnotify:= GetTickCount
          end
    end;  { i loop over paragraphs }

  currentdyn.DynStyle:= 0;
  runningcolor:= clNone;
  HtmlFormat([], currentdyn, runningcolor);
  StreamOut('');
  StreamOut('</DIV>');
  if full then
    begin
      if BodyTag then StreamOut('</BODY>');
      if HeadTag then StreamOut('</HTML>')
    end;

  if accumulator<>nil then StrDispose(accumulator)
end;

function TPlusToHTML.SizeToString(Size: Integer): string;
begin
  if Assigned(OnGetSize) then OnGetSize(Self, Size, Result)
  else
    if Size<5 then Result:= '0'
    else
      if Size<18 then Result:= IntToStr((Abs(Size)-10) div 2 + 3)
                 else Result:= '7';
  Result:= '"'+Result+'"'
end;

var i: Integer;

initialization
  RTFClipboardFormat:= 0;  { mark them as not registered }
  HtmlClipboardFormat:= 0;

  SpecialChars:= [];
  for i:= 0 to SpecCharCount-1 do
    Include(SpecialChars, SpecChars[i].c)

end.

